/*
 *
 * Pattern 2 0.4
 * Based on AS3DFile.c from 
 *  AS3D Engine 4.14
 * and pattern.c from rockband2-drum-0.1
 *	by Joel R. Voss
 *      06/01/2010 - 06/10/2010
 *
 */

#include <stdio.h>
#include <memory.h>
#include <stdlib.h>
#include <string.h>

#include "pattern2.h"

/**
 * pattern_AS3DFile()
 * A required initializer based on AS3DFile 4.14.
 */
int pattern_AS3DFile(pattern *p, int iVerbosity, char *sHeader)
{
	if(iVerbosity > 0)
	{
		p->m_bErrors  = 1;
		p->m_bVerbose = 1;
		p->m_bDebug   = (iVerbosity >= 2);
	}
	else
	{
		p->m_bErrors = (iVerbosity >= 0);
		p->m_bVerbose = 0;
		p->m_bDebug   = 0;
	}
	p->m_sAppHdr = sHeader;
	p->m_sTitle = (char *)malloc(BUFFER_SIZE);
	p->m_sScene = (char *)malloc(BUFFER_SIZE);
	p->m_iScene = 0;

	p->firstChild  = 0;
	p->nextSibling = 0;
	p->value       = 0;
	p->value_len   = 0;

	p->output      = 0;
	p->output_len  = 0;
	
	//memset(&(p->m_pUInput), 0, sizeof(p->m_pUInput));

	// dne p->m_iWidth = 0;
	return 1;
}

/**
 * pattern_deconst_AS3DFile()
 * A required deinitializer based on AS3DFile 4.14.
 */
int pattern_deconst_AS3DFile(pattern *p)
{
	// debug if(m_bDebug) cout << "end of file" << endl;
	// debug if(m_bDebug) cout << "SAFE_DELETE_ARRAY(m_pLights)" << endl;
	// dne SAFE_DELETE_ARRAY(m_pLights);
	if (p->m_sTitle)
	{
		free(p->m_sTitle);
		p->m_sTitle = 0;
	}
	if (p->m_sScene)
	{
		free(p->m_sScene);
		p->m_sScene = 0;
	}
	if (p->value)
	{
		free(p->value);
		p->value = 0;
		p->value_len = 0;
	}
	if (p->output)
	{
		free(p->output);
		p->output = 0;
		p->output_len = 0;
	}
	if (p->firstChild)
	{
		pattern_deconst_AS3DFile(p->firstChild);
		free(p->firstChild);
		p->firstChild = 0;
	}
	if (p->nextSibling)
	{
		pattern_deconst_AS3DFile(p->nextSibling);
		free(p->nextSibling);
		p->nextSibling = 0;
	}

	return 1;
}

/* pattern_destroy() is a temporary delete.
 * If you need to ditch all the renderable content quickly
 * and then reload it onto another context, this is the way to go.
 */
int pattern_destroy(pattern *p)
{

	// debug if(m_bDebug) cout << "destroying file" << endl;

	// debug if(m_bDebug) cout << "SAFE_DELETE_ARRAY(m_pVel)" << endl;
	// dne SAFE_DELETE_ARRAY(m_pVel);
	
	if (p->m_sTitle)
	{
		free(p->m_sTitle);
		p->m_sTitle = 0;
	}
	if (p->m_sScene)
	{
		free(p->m_sScene);
		p->m_sScene = 0;
	}
	if (p->value)
	{
		free(p->value);
		p->value = 0;
		p->value_len = 0;
	}
	if (p->output)
	{
		free(p->output);
		p->output = 0;
		p->output_len = 0;
	}
	if (p->firstChild)
	{
		pattern_destroy(p->firstChild);
		free(p->firstChild);
		p->firstChild = 0;
	}
	if (p->nextSibling)
	{
		pattern_destroy(p->nextSibling);
		free(p->nextSibling);
		p->nextSibling = 0;
	}

	return 1;
}

/**
 * pattern_addChild()
 * A tree-based pattern child adder. Child must be malloc'ed.
 */
int pattern_addChild(pattern *p, pattern *child)
{
	pattern **myChild = &(p->firstChild);
	while(*myChild != 0)
	{
		// TODO: Am I doing this right?
		myChild = &((*myChild)->nextSibling);
	}
	*myChild = child;
	return 1;
}

/**
 * pattern_addSibling()
 * A tree-based pattern sibling adder. Sibling must be malloc'ed.
 */
int pattern_addSibling(pattern *p, pattern *sibling)
{
	pattern **mySibling = &(p->nextSibling);
	while(*mySibling != 0)
	{
		// TODO: Am I doing this right?
		mySibling = &((*mySibling)->nextSibling);
	}
	*mySibling = sibling;
	return 1;
}

/**
 * pattern_load()
 * A tree-based pattern file loader. Pattern must be malloc'ed.
 */
int pattern_load(pattern *p, const char *sFilename)
{
	FILE *asd_in = fopen(sFilename, "r");
	if (asd_in == 0)
	{
		if(p->m_bErrors) printf("Error opening file.\n");
		return 0;
	}

	int bRet = pattern_readData(p, asd_in);

	fclose(asd_in);

	return bRet;
}

/**
 * pattern_readData()
 * A tree-based pattern file reader. Pattern must be malloc'ed.
 */
int pattern_readData(pattern *p, FILE *asd_in)
{
	// Brilliant idea: chuck buffer to getNextLine(),
	// which will return sLine. Sweet, huh?
	////char buffer[]; no use.
	char sLine[BUFFER_SIZE];
	char sTempHdr[BUFFER_SIZE];
	int iDelimPos = -1;

	// Header
	if(getNextLine(asd_in, sLine) == 0) return 0;
	if(strcmp(sLine, p->m_sAppHdr) != 0)
	{
		if(p->m_bErrors) writeError(sLine, p->m_sAppHdr);
		return 0;
	}
	else
	{
		// AS3D Manga Director Version 4.0-0.9
		if(p->m_bVerbose) printf("Got header correctly.\n%s\n", sLine);
	}

	iDelimPos = -1;
	if(getNextLine(asd_in, sLine) == 0) return 0;
	getNextWord(sTempHdr, sLine, &iDelimPos);
	if(strcmp(AS3D_TITLE, sTempHdr) != 0)
	{
		if(p->m_bErrors) writeError(sTempHdr, AS3D_TITLE);
		return 0;
	}

	if(p->m_sTitle == 0)
	{
		// Memory error...
		return 0;
	}
	// *Title: Javantea's Fate
	// TODO: Deal with pointer math...
	if(sLine[iDelimPos] == ' ')
	{
		iDelimPos++;
	}
	// TODO: Deal with strncpy
	strncpy(p->m_sTitle, sLine + iDelimPos, BUFFER_SIZE);
	
	/*
	if(initHeader(asd_in) == 0)
	{
		if(p->m_bErrors) printf("Error with Header;\n");
		return 0;
	}
	*/

	iDelimPos = -1;
	if(getNextLine(asd_in, sLine) == 0) return 0;
	getNextWord(sTempHdr, sLine, &iDelimPos);
	if(strcmp(AS3D_SCENE, sTempHdr) != 0)
	{
		if(p->m_bErrors) writeError(sTempHdr, AS3D_SCENE);
		return 0;
	}

	p->m_iScene = getNextInt(sLine, &iDelimPos);
	//printf("haha %i\n", p->m_iScene);

	if(p->m_sScene == 0)
	{
		// Memory error...
		return 0;
	}
	//printf("tbp: %s\n", sLine+iDelimPos);
	getNextWord(p->m_sScene, sLine, &iDelimPos);
	//printf("ah %s\n", p->m_sScene);

	// Children and siblings to be parsed here...
	while(1)
	{
		if(getNextLine(asd_in, sLine) == 0) return 0;
	
		if(strcmp(sLine, AS3D_CHILD) == 0)
		{
			pattern *child = (pattern *)malloc(sizeof(pattern));
			pattern_AS3DFile(child, 2, PATTERN2_HEADER);
			pattern_readData(child, asd_in);
			pattern_addChild(p, child);
		}
		if(strcmp(sLine, AS3D_SIBLING) == 0)
		{
			pattern *sibling = (pattern *)malloc(sizeof(pattern));
			pattern_AS3DFile(sibling, 2, PATTERN2_HEADER);
			pattern_readData(sibling, asd_in);
			pattern_addSibling(p, sibling);
		}
		if(strcmp(sLine, AS3D_END) == 0)
		{
			return 1;
		}
		if (feof(asd_in))
		{
			if(p->m_bErrors) writeError(sLine, AS3D_END);
			return 0;
		}
	}
	return 1;

}

/**
 * pattern_saveASD()
 * A tree-based pattern file writer. Pattern must be malloc'ed.
 */
int pattern_saveASD(pattern *p, const char *filename)   // Save data to data file
{
	int bRet = 1;
	FILE *asd_out = fopen(filename, "w");
	if (!asd_out)
	{
		if(p->m_bErrors) printf("Error opening file.\n");
		return 0;
	}
	
	bRet = pattern_writeData(p, asd_out);

	fclose(asd_out);
	
	return bRet;
}

/**
 * pattern_writeData()
 * A tree-based pattern file writer. Pattern must be malloc'ed.
 */
int pattern_writeData(pattern *p, FILE *asd_out)   // Save data to data file
{
	int bRet = 1;
	
	// AS3D Manga Director Version 4.0-0.8
	fprintf(asd_out, "%s\n", p->m_sAppHdr);
	// *Title: Javantea's Fate
	fprintf(asd_out, "%s %s\n", AS3D_TITLE, p->m_sTitle);
	// *Scene: 1 "Warehouse"
	fprintf(asd_out, "%s %i %s\n", AS3D_SCENE, p->m_iScene, p->m_sScene);

	uint32_t tmp = 0;
	if(p->value && p->value_len > 0)
	{
		// If it's >= 32-bits, display 32-bits.
		if(p->value_len >= 4)
			tmp = *((uint32_t *)p->value);
		else
			// Otherwise show 8-bits.
			tmp = *p->value;
	}
	fprintf(asd_out, "%s %08x %c\n", AS3D_VALUE, tmp, (char)tmp);

	if(p->firstChild)
	{
		fprintf(asd_out, "Child:\n");
		bRet = pattern_writeData(p->firstChild, asd_out);
	}
	
	// TODO: handle bRet before it gets overwritten.

	if(p->nextSibling)
	{
		fprintf(asd_out, "Sibling:\n");
		bRet = pattern_writeData(p->nextSibling, asd_out);
	}
	
	// *Options: windowed 1024 768 24
	fprintf(asd_out, "%s\n", AS3D_END);

	return bRet;
}

/**
 * pattern_setDir()
 * A function that does nothing currently. It was used for AS3DFile so that 
 * conf and data files in the program's directory could be found. That is no 
 * longer necessary in pattern2.
 */
int pattern_setDir(pattern *p, char *aDir)
{
	if(aDir == 0) return 0;
	/*
	int bRet = 1;
	string sDir(aDir);
	int iEnd = sDir.find_last_of("/");
	if(iEnd<0)
	{
		m_sDir = "";
		return 0;
	}
	if(m_bVerbose) cout << "slash is found at: " << iEnd << "." << endl;
	m_sDir = sDir.substr(0, iEnd+1);
	if(m_bVerbose) cout << "'" << sDir << "' gives us '" << m_sDir << "';" << endl;
	return bRet;
	// Hopefully at a later date, I can get this so that the loaders (ms3d/img) will
	// recognize the binary directory.
	*/
	return 1;
}

/**
 * pattern_compare()
 * From pattern.c, this code is strange and awesome.
 * It takes a tree and a list and compares them.
 * It returns the number of matches and sets the match argument.
 */
int pattern_compare(pattern *p, pattern *list, pattern **match)
{
	pattern *found = 0;
	pattern *prev_found = 0;
	int found_items = 0;
	pattern *current_list = list;
	pattern *current_tree = p;
	// Calculate how long the match is.
	int match_len = 0;
	while((current_list != 0) && (current_tree != 0))
	{
		//printf("compare: %i %i\n", current_tree->value, current_list->value);
		found = pattern_find(current_tree, current_list->value, 1);
		//printf("f %i\n", match_len);
		if(!found)
		{
			if(prev_found)
			{
				found = prev_found;
			}
			// Not found.
			break;
		}
		prev_found = found;
		match_len++;
		// Lists are made of children, not siblings.
		current_list = current_list->firstChild;
		current_tree = found->firstChild;
		//printf("%p %p\n", current_list, current_tree);
	}
	// TODO: Also tell the user how many matches occur.
	if(found)
	{
		found_items = 1;
		if(found->firstChild)
		{
			found_items++;
			if (found->firstChild->nextSibling || found->firstChild->firstChild)
				found_items++;
		}
	}
	*match = found;
	return match_len;
}

/**
 * pattern_find()
 * From pattern.c.
 * It takes a tree and a string and compares each sibling.
 * Find just on this level.
 * It returns a leaf of the tree.
 */
pattern *pattern_find(pattern *p, uint8_t *value, int len)
{
	if(p == 0)
		return 0;
	pattern *ret = 0;
	pattern *current = p;
	while(current != 0)
	{
		//printf("%s ?= %s, %i %i\n", current->value, value, len, current->value_len);
		//printf("%p\n", current->value);
		if(current->value)
		{
			//printf("haha %hhi %hhi\n", current->value[0], value[0]);
			if(memcmp(current->value, value, min(len, current->value_len)) == 0)
				return current;
		}
		current = current->nextSibling;
	}
	return ret;
}

/**
 * pattern_input()
 * Add to a tree with a string and an output string.
 * Returns 0 on success, 1 on failure.
 */
int pattern_input(pattern *tree, uint8_t *tree_input, int len, uint8_t *out_value, int out_len)
{
	uint8_t *input_curr = tree_input;
	pattern *prev_tree = tree;
	int i;
	for(i = 0; i < len; i++)
	{
		pattern *t = pattern_find(prev_tree->firstChild, input_curr, 1);
		if(t == 0)
		{
			t = (pattern *)malloc(sizeof(pattern));
			pattern_AS3DFile(t, 2, AS3D_TITLE);
			// TODO: write a function that does a memcpy.
			t->value = malloc(2);
			t->value[0] = input_curr[0];
			t->value[1] = 0;
			t->value_len = 1;
			//printf("pattern_tree sizeof(%i), buf.pos %i\n", sizeof(pattern_tree), buf->pos);
			if(t == 0)
			{
				printf("Error with malloc pattern.\n");
				return 1;
			}
			pattern_addChild(prev_tree, t);
		}
		else
		{
			//printf("found tree?\n");
		}
		prev_tree = t;
		input_curr++;
	}
	if(!prev_tree)
	{
		printf("without a prev_tree, pattern_input failed\n");
		return 1;
	}
	if(out_len > 0 && out_value != 0)
	{
		prev_tree->output = malloc(out_len);
		memcpy(prev_tree->output, out_value, out_len);
		prev_tree->output_len = out_len;
	}
	return 0;
}



#ifdef TEST_PATTERN

/**
 * p1Test0(argc, argv)
 * A test func which loads a tree and then saves it to a file.
 * 
 */
int p1Test0(int argc, char **argv)
{
	// All the below stuff works great.
	pattern p;
	int r;
	
	const char *filename = "pattern2a1.txt";
	if(argc > 1)
	{
		filename = argv[1];
	}
	
	// 2 == debug
	r = pattern_AS3DFile(&p, 2, PATTERN2_HEADER);
	
	if(r == 0)
	{
		printf("Sorry, File init failed.\n");
		return 1;
	}
	
	r = pattern_load(&p, filename);
	
	if(r == 0)
	{
		printf("Sorry, File load failed.\n");
		return 1;
	}


	r = pattern_saveASD(&p, "pattern2a2.txt");
	
	if(r == 0)
	{
		printf("Sorry, File save failed.\n");
		return 1;
	}
	
	r  = pattern_destroy(&p);
	r += pattern_deconst_AS3DFile(&p);
	
	if(r == 2)
	{
		printf("Complete success\n");
	}
	else
	{
		printf("Sorry, Deconstruction failed.\n");
	}
	
	return 0;
}

/**
 * p1Test(p)
 * A test func which loads a static conf from memory and then writes the trees 
 * to two files. The input pattern is not currently used.
 * 
 */
int p1Test(pattern *p)
{
	// The list.
	pattern *l = (pattern *)malloc(sizeof(pattern));
	pattern_AS3DFile(l, 2, PATTERN2_HEADER);
	// Create a tree, length N, data blahyakdaah, ...
	uint8_t *tree_input = (uint8_t *)"blahyakdaah";
	pattern *root_tree = (pattern *)malloc(sizeof(pattern));
	pattern_AS3DFile(root_tree, 2, PATTERN2_HEADER);
	int ret = pattern_input(root_tree, tree_input, 11, 0, 0);
	ret = pattern_input(root_tree, (uint8_t *)"blahdah", 7, 0, 0);
	ret = pattern_input(root_tree, (uint8_t *)"beee", 4, 0, 0);
	ret = pattern_input(root_tree, (uint8_t *)"befg", 4, 0, 0);
	ret = pattern_input(root_tree, (uint8_t *)"abcd", 4, 0, 0);
	
	//pattern *next = (pattern *)malloc(sizeof(pattern));
	ret = pattern_input(l, (uint8_t *)"befg", 4, 0, 0);
	
	pattern *found;
	int match_len = pattern_compare(root_tree->firstChild, l->firstChild, &found);
	printf("match length: %i\n", match_len);
	if(match_len > 0 && found)
	{
		// This should contain befg, not g... or maybe...
		printf("data: %hhi\n", found->value[0]);
	}
	
	pattern_saveASD(l, "pattern2cl.txt");
	pattern_saveASD(root_tree, "pattern2ct.txt");
	
	pattern_deconst_AS3DFile(l);
	pattern_deconst_AS3DFile(root_tree);
	
	return 1;
}


/**
 * p1Test2()
 * A test func which loads a static conf from memory and then expects data on 
 * the cli.
 * 
 */
int p1Test2()
{
	// Create a tree, length N, data blahyakdaah, ...
	uint8_t *tree_input = (uint8_t *)"blahyakdaah";
	pattern *root_tree = (pattern *)malloc(sizeof(pattern));
	pattern_AS3DFile(root_tree, 2, PATTERN2_HEADER);
	int ret = pattern_input(root_tree, tree_input, 11, (uint8_t *)"b4c4", 4);
	ret = pattern_input(root_tree, (uint8_t *)"blahdah", 7, (uint8_t *)"test", 4);
	ret = pattern_input(root_tree, (uint8_t *)"beee", 4, (uint8_t *)"been", 4);
	ret = pattern_input(root_tree, (uint8_t *)"befg", 4, (uint8_t *)"beff", 4);
	ret = pattern_input(root_tree, (uint8_t *)"abcd", 4, (uint8_t *)"areyousure?", 11);
	
	int data_len = 1;
	char data[256];
	while(!feof(stdin))
	{
		if(!fgets(data, 256, stdin))
		{
			printf("End of File\n");
			break;
		}
		data[255] = 0;
		data_len = strlen(data) - 1;
		//printf("(%i) %s\n", data_len, data);
		if(data_len < 1)
		{
			// Don't parse
			continue;
		}
		// The list.
		pattern *l = (pattern *)malloc(sizeof(pattern));
		pattern_AS3DFile(l, 2, PATTERN2_HEADER);
		//pattern *next = (pattern *)malloc(sizeof(pattern));
		ret = pattern_input(l, (uint8_t *)data, data_len, 0, 0);
		//printf("%p, %p, %p\n", l, l->firstChild, l->firstChild->nextSibling);
		
		pattern *found;
		int match_len = pattern_compare(root_tree->firstChild, l->firstChild, &found);
		printf("match length: %i\n", match_len);
		if(match_len > 0 && found)
		{
			// This should contain befg, not g... or maybe...
			printf("data: %c\n", found->value[0]);
			if(found->output && found->output_len)
			{
				printf("output ");
				ret = fwrite(found->output, found->output_len, 1, stdout);
				printf("\n");
			}
		}
		pattern_deconst_AS3DFile(l);
	}
	
	pattern_deconst_AS3DFile(root_tree);
	
	return 1;
}

/**
 * p2main(p2conf_filename)
 * The main func which loads a conf file and then expects data on the cli.
 * 
 */
int p2main(char *p2conf_filename)
{
	int ret = 0;
	// Create a tree, length N, data blahyakdaah, ...
	pattern *root_tree = (pattern *)malloc(sizeof(pattern));
	pattern_AS3DFile(root_tree, 2, PATTERN2_HEADER);
	FILE *pattern_conf = fopen(p2conf_filename, "r");
	if(!pattern_conf)
	{
		printf("Error opening %s\n", p2conf_filename);
		return 1;
	}
	while(!feof(pattern_conf))
	{
		uint8_t tree_input[BUFFER_SIZE];
		int tree_input_len;
		uint8_t tree_output[BUFFER_SIZE];
		int tree_output_len;
		tree_input[0]  = 0;
		tree_output[0] = 0;
		char sLine[BUFFER_SIZE];
		// Require a line, otherwise we're done.
		if(getNextLine(pattern_conf, sLine) == 0) break;
		//if(sLine[0] == 0) break;
		int iDelimPos = -1;
		getNextWord((char *)tree_input,  sLine, &iDelimPos);
		getNextWord((char *)tree_output, sLine, &iDelimPos);
		tree_input_len  = strlen((char *)tree_input);
		tree_output_len = strlen((char *)tree_output);
		printf("%s, %s, %i, %i\n", tree_input, tree_output, tree_input_len, tree_output_len);
		ret = pattern_input(root_tree, tree_input, tree_input_len, tree_output, tree_output_len);
	}
	fclose(pattern_conf);
	int data_len = 1;
	char data[256];
	while(!feof(stdin))
	{
		if(!fgets(data, 256, stdin))
		{
			printf("End of File\n");
			break;
		}
		data[255] = 0;
		data_len = strlen(data) - 1;
		//printf("(%i) %s\n", data_len, data);
		if(data_len < 1)
		{
			// Don't parse
			continue;
		}
		// The list.
		pattern *l = (pattern *)malloc(sizeof(pattern));
		pattern_AS3DFile(l, 2, PATTERN2_HEADER);
		//pattern *next = (pattern *)malloc(sizeof(pattern));
		ret = pattern_input(l, (uint8_t *)data, data_len, 0, 0);
		//printf("%p, %p, %p\n", l, l->firstChild, l->firstChild->nextSibling);
		
		pattern *found;
		int match_len = pattern_compare(root_tree->firstChild, l->firstChild, &found);
		printf("match length: %i\n", match_len);
		if(match_len > 0 && found)
		{
			// This should contain befg, not g... or maybe...
			printf("data: %c\n", found->value[0]);
			if(found->output && found->output_len)
			{
				printf("output ");
				ret = fwrite(found->output, found->output_len, 1, stdout);
				printf("\n");
			}
		}
		pattern_deconst_AS3DFile(l);
	}
	
	pattern_deconst_AS3DFile(root_tree);
	
	return 1;
}

int main(int argc, char **argv)
{
	if(argc > 1 && (strcmp(argv[1], "-h") == 0))
	{
		printf("Usage: pattern2 [p2.conf]\n");
		return 1;
	}
	char *filename = "p2.conf";
	if(argc > 1)
	{
		filename = argv[1];
	}
	p2main(filename);
	
	return 0;
}

#endif // TEST_PATTERN
